# from genet.utils import __version__

'''
genet의 처음 init 부분
'''

# How to release with 2FA accounts

# conda activate py38
# hatch build
# twine upload --username __token__ dist/*
# [Enter PyPI API token]

__version__ = '0.12.0'